
#define NEED_EVENTS
#define NEED_REPLIES

#include <X11/X.h>
#include <X11/Xproto.h>
#include "misc.h"
#include "os.h"
#include "dixstruct.h"
#include "extnsionst.h"
#include "scrnintstr.h"

#include <tslib.h>

#include <X11/extensions/xspproto.h>
#include <X11/extensions/xspwire.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>

#include "spext.h"

#ifdef EXTMODULE
#include "xf86_ansic.h"
#endif

#define XSP_EXTENSION_NAME "XSP"
#define XSP_NUMBER_EVENTS  XSPNumberEvents
#define XSP_NUMBER_ERRORS  XSPNumberErrors

#define XSP_EXTENSION_MAJOR         1
#define XSP_EXTENSION_MINOR         0
#define XSP_EXTENSION_PATCHLEVEL    1


/*
 * for /dev/dspctl/ctl
 */

#define OMAP_DSP_IOCTL_FBEN                     53
#define OMAP_DSP_IOCTL_FBDIS                    54


static unsigned char    XSPReqCode;
int                     XSPEventBase;
int                     XSPReqBase;
int                     XSPErrorBase;

int xspScrPrivateIndex = -1;
static int XSPGeneration = -1;

static ClientPtr xsp_ts_client;
static void XSPClientCallback (CallbackListPtr *list,
        pointer closure,
        pointer data);


extern void (*tslib_raw_event_hook)(int x, int y, int pressure, void *closure);
extern void *tslib_raw_event_closure;
void (*xsp_dsp_event_hook)(void);

int xsp_init_screen(ScreenPtr pScreen);


/* from hw/kdrive/src/kdrive.h */
void KdDisableInput (void);
void KdEnableInput (void);


static DISPATCH_PROC(ProcXSPDispatch);
static DISPATCH_PROC(SProcXSPDispatch);
static void ProcXSPCloseDown(ExtensionEntry *);

static DISPATCH_PROC(ProcXSPQueryVersion);
static DISPATCH_PROC(ProcXSPSetTSRawMode);
static DISPATCH_PROC(ProcXSPSetTSCalibration);
static DISPATCH_PROC(ProcXSPRegisterDSPArea);
static DISPATCH_PROC(ProcXSPCancelDSPArea);
static DISPATCH_PROC(ProcXSPSetPixelDoubling);

static DISPATCH_PROC(SProcXSPQueryVersion);
static DISPATCH_PROC(SProcXSPSetTSRawMode);
static DISPATCH_PROC(SProcXSPSetTSCalibration);
static DISPATCH_PROC(SProcXSPRegisterDSPArea);
static DISPATCH_PROC(SProcXSPCancelDSPArea);
static DISPATCH_PROC(SProcXSPSetPixelDoubling);

void ProcXSPCloseDown (ExtensionEntry*);

static void xsp_ts_event_hook (int x, int y, int pressure, void *closure)
{
    ClientPtr pClient = (ClientPtr) closure;
    xXSPRawTouchscreenEvent        ev;

    ev.type = XSPEventBase + X_XSPTSRaw;
    ev.sequenceNumber = pClient->sequence;
    ev.x = x;
    ev.y = y;
    ev.pressure = pressure;

    if (!pClient->clientGone)
        WriteEventsToClient (pClient, 1, (xEvent *) &ev);
}


int xsp_init_screen(ScreenPtr pScreen)
{
    xspScrPrivPtr pScrPriv;

    if (XSPGeneration != serverGeneration) {
        if ((xspScrPrivateIndex = AllocateScreenPrivateIndex()) == -1) {
            FatalError("XSPExtensionInit: AllocateScreenPrivateIndex failed");
            return FALSE;
        }

        XSPGeneration = serverGeneration;
    }


    if (pScreen->devPrivates[xspScrPrivateIndex].ptr) {
        return TRUE; /* already allocated */
    }

    pScrPriv = (xspScrPrivPtr)xalloc(sizeof (xspScrPrivRec));
    if (!pScrPriv) {
        return FALSE;
    }

    memset(pScrPriv, '\0', sizeof(xspScrPrivRec));
    pScreen->devPrivates[xspScrPrivateIndex].ptr = (pointer)pScrPriv;

    return TRUE;
}

void XSPExtensionInit(void)
{
    ExtensionEntry *extEntry;
    int s;

    if (!AddCallback(&ClientStateCallback, XSPClientCallback, 0))
        return;

    extEntry = AddExtension(XSP_EXTENSION_NAME,
            XSP_NUMBER_EVENTS,
            XSP_NUMBER_ERRORS,
            ProcXSPDispatch, SProcXSPDispatch,
            ProcXSPCloseDown,
            StandardMinorOpcode);
    if (!extEntry)
    {
        FatalError("XSPExtensionInit: AddExtension failed");
        return;
    }

    for (s=0; s < screenInfo.numScreens; s++) {
        if (!xsp_init_screen(screenInfo.screens[s])) {
            FatalError("XSP :: xsp_init_screen failed");
            return;
        }
    }
    XSPReqCode = (unsigned char)extEntry->base;
    fprintf(stderr, "XSPReqCode: %d\n", XSPReqCode);
    XSPEventBase = extEntry->eventBase;
    XSPErrorBase = extEntry->errorBase;
    xsp_ts_client = 0;
}

static int ProcXSPDispatch(client)
    register ClientPtr	client;
{
    REQUEST(xReq);
    fprintf(stderr, "ProcXSPDispatch: %d\n", stuff->data);
    switch (stuff->data)
    {
        case X_XSPQueryVersion:
            return ProcXSPQueryVersion(client);
        case X_XSPSetTSCalibration:
            return ProcXSPSetTSCalibration(client);
        case X_XSPSetTSRawMode:
            return ProcXSPSetTSRawMode(client);
        case X_XSPRegisterDSPArea:
            return ProcXSPRegisterDSPArea(client);
        case X_XSPCancelDSPArea:
            return ProcXSPCancelDSPArea(client);
        case X_XSPSetPixelDoubling:
            return ProcXSPSetPixelDoubling(client);
        default:
            return BadRequest;
    }
}

static int SProcXSPDispatch(ClientPtr client)
{
    REQUEST(xReq);
    fprintf(stderr, "SProcXSPDispatch: %d\n", stuff->data);
    switch (stuff->data)
    {
        case X_XSPQueryVersion:
            return SProcXSPQueryVersion(client);
        case X_XSPSetTSCalibration:
            return SProcXSPSetTSCalibration(client);
        case X_XSPSetTSRawMode:
            return SProcXSPSetTSRawMode(client);
        case X_XSPRegisterDSPArea:
            return SProcXSPRegisterDSPArea(client);
        case X_XSPCancelDSPArea:
            return SProcXSPCancelDSPArea(client);
        default:
            fprintf(stderr, "BadRequest: %d\n", stuff->data);
            return BadRequest;
    }
}


static int ProcXSPQueryVersion(client)
    register ClientPtr	client;
{
    xXSPQueryVersionReply rep;
    register int n;

    REQUEST_SIZE_MATCH(xXSPQueryVersionReq);
    rep.type = X_Reply;
    rep.length = 0;
    rep.sequenceNumber = client->sequence;
    rep.majorVersion = XSP_VERSION_MAJOR;
    rep.minorVersion = XSP_VERSION_MINOR;
    if (client->swapped) {
        swaps(&rep.sequenceNumber, n);
        swaps(&rep.minorVersion, n);
    }
    WriteToClient(client, sizeof(xXSPQueryVersionReply), (char *)&rep);
    return(client->noClientException);
}

static int SProcXSPQueryVersion(ClientPtr client)
{
    register int n;
    REQUEST(xXSPQueryVersionReq);

    swaps(&stuff->length, n);
    REQUEST_SIZE_MATCH(xXSPQueryVersionReq);
    swaps(&stuff->minorVersion, n);
    return ProcXSPQueryVersion(client);
}





static int ProcXSPSetTSRawMode (ClientPtr client)
{
    REQUEST(xXSPSetTSRawModeReq);
    xXSPSetTSRawModeReply rep;

    REQUEST_SIZE_MATCH (xXSPSetTSRawModeReq);

    memset (&rep, 0, sizeof (rep));
    rep.type = X_Reply;
    rep.sequenceNumber = client->sequence;

    if (stuff->on)
    {
        if (xsp_ts_client == NULL)
        {
            /* Start calibrating.  */
            xsp_ts_client = client;
            tslib_raw_event_hook = xsp_ts_event_hook;
            tslib_raw_event_closure = client;
            rep.status = GrabSuccess;
        }
        else
        {
            rep.status = AlreadyGrabbed;
        }
    } else {
        if (xsp_ts_client == client)
        {
            /* Stop calibrating.  */
            xsp_ts_client = NULL;
            tslib_raw_event_hook = NULL;
            tslib_raw_event_closure = NULL;
            rep.status = GrabSuccess;
        } else {
            rep.status = AlreadyGrabbed;
        }
    }

    if (client->swapped) {
        int n;

        swaps (&rep.sequenceNumber, n);
        swaps (&rep.status, n);
    }
    WriteToClient(client, sizeof (rep), (char *) &rep);
    return (client->noClientException);
}

static int SProcXSPSetTSRawMode (ClientPtr client)
{
    REQUEST(xXSPSetTSRawModeReq);
    int n;

    REQUEST_SIZE_MATCH (xXSPSetTSRawModeReq);

    swaps(&stuff->on, n);

    return ProcXSPSetTSRawMode(client);
}

static int ProcXSPSetTSCalibration(ClientPtr client)
{
    REQUEST(xXSPSetTSCalibrationReq);
    xXSPSetTSCalibrationReply rep;

    REQUEST_SIZE_MATCH (xXSPSetTSCalibrationReq);

    memset (&rep, 0, sizeof (rep));
    rep.type = X_Reply;
    rep.sequenceNumber = client->sequence;

    fprintf(stderr, "new calibration constants: %i, %i, %i, %i, %i, %i, %i\n",
            (int)stuff->calconst1,
            (int)stuff->calconst2,
            (int)stuff->calconst3,
            (int)stuff->calconst4,
            (int)stuff->calconst5,
            (int)stuff->calconst6,
            (int)stuff->calconst7);

    if (ts_write_calibration(stuff->calconst1,
                stuff->calconst2,
                stuff->calconst3,
                stuff->calconst4,
                stuff->calconst5,
                stuff->calconst6,
                stuff->calconst7) < 0) {
        rep.status = 0;
    } else {
        rep.status = 1;
    }

    /* cycle input to take new values into use */
    KdDisableInput();
    KdEnableInput();

    if (client->swapped)
    {
        int n;

        swaps (&rep.sequenceNumber, n);
        swaps (&rep.status, n);
    }
    WriteToClient(client, sizeof (rep), (char *) &rep);
    return (client->noClientException);
}


static int SProcXSPSetTSCalibration(ClientPtr client)
{
    REQUEST(xXSPSetTSCalibrationReq);
    register int n;
    REQUEST_SIZE_MATCH(xXSPSetTSCalibrationReq);

    swaps(&stuff->calconst1, n);
    swaps(&stuff->calconst2, n);
    swaps(&stuff->calconst3, n);
    swaps(&stuff->calconst4, n);
    swaps(&stuff->calconst5, n);
    swaps(&stuff->calconst6, n);
    swaps(&stuff->calconst7, n);

    return ProcXSPSetTSCalibration(client);
}




static int ProcXSPRegisterDSPArea(ClientPtr client)
{

    REQUEST(xXSPRegisterDSPAreaReq);
    xXSPRegisterDSPAreaReply rep;
    register int n;
    int dsp_ctl_fd;
    REQUEST_SIZE_MATCH(xXSPRegisterDSPAreaReq);

    ScreenPtr scr;
    xspScrPrivPtr pScrPriv;

    fprintf(stderr, "registering dsp area\n");
    memset(&rep, 0, sizeof(rep));
    rep.type = X_Reply;
    rep.sequenceNumber = client->sequence;
    rep.status = 0;

    fprintf(stderr, "register dsp area: %i, (%i, %i), %i, %i\n",
            (int)stuff->screen_num, (int)stuff->x, (int)stuff->y,
            (int)stuff->width, (int)stuff->height);


    if (stuff->screen_num < 0 || stuff->screen_num > (screenInfo.numScreens - 1)) {
        rep.status = 0;
        goto getout;
    }

    scr = screenInfo.screens[stuff->screen_num];
    pScrPriv = (pointer)scr->devPrivates[xspScrPrivateIndex].ptr;
    pScrPriv->dsp_client = client;
    pScrPriv->dsp_enabled = 1;
    pScrPriv->dsp_box.x1 = stuff->x;
    pScrPriv->dsp_box.y1 = stuff->y;
    pScrPriv->dsp_box.x2 = stuff->x + stuff->width;
    pScrPriv->dsp_box.y2 = stuff->y + stuff->height;

    if ( (dsp_ctl_fd = open("/dev/dspctl/ctl", O_RDWR)) < 0) {
        perror("Xomap: Unable to open /dev/dspctl/ctl to stop DSP\n");
    } else {
        if (ioctl(dsp_ctl_fd, OMAP_DSP_IOCTL_FBEN) < 0) {
            perror("Xomap: ioctl OMAP_DSP_IOCTL_FBDBDIS failed");
        }
        close(dsp_ctl_fd);
    }

    rep.status = 1;

getout:
    if (client->swapped) {
        swaps(&rep.sequenceNumber, n);
        swaps(&rep.status, n);
    }

    WriteToClient(client, sizeof(rep), (char *)&rep);
    return (client->noClientException);
}

static int SProcXSPRegisterDSPArea(ClientPtr client)
{
    REQUEST(xXSPRegisterDSPAreaReq);
    register int n;
    REQUEST_SIZE_MATCH(xXSPRegisterDSPAreaReq);
    swaps(&stuff->screen_num, n);
    swaps(&stuff->x, n);
    swaps(&stuff->y, n);
    swaps(&stuff->width, n);
    swaps(&stuff->height, n);

    return ProcXSPRegisterDSPArea(client);
}




static int ProcXSPCancelDSPArea(ClientPtr client)
{
    REQUEST(xXSPCancelDSPAreaReq);
    xXSPCancelDSPAreaReply rep;
    int dsp_ctl_fd;
    REQUEST_SIZE_MATCH(xXSPCancelDSPAreaReq);
    xspScrPrivPtr pScrPriv;

    memset(&rep, 0, sizeof(rep));
    rep.type = X_Reply;
    rep.sequenceNumber = client->sequence;

    if (stuff->screen_num < 0 || stuff->screen_num > (screenInfo.numScreens - 1)) {
        rep.status = 0;
        goto getout;
    }

    pScrPriv = screenInfo.screens[stuff->screen_num]->devPrivates[xspScrPrivateIndex].ptr;
    pScrPriv->dsp_client = NULL;
    pScrPriv->dsp_enabled = 0;
    pScrPriv->dsp_box.x1 = 29999;
    pScrPriv->dsp_box.y1 = 29999;
    pScrPriv->dsp_box.x2 = -29999;
    pScrPriv->dsp_box.y2 = -29999;
    if ( (dsp_ctl_fd = open("/dev/dspctl/ctl", O_RDWR)) < 0) {
        perror("Xomap: Unable to open /dev/dspctl/ctl to stop DSP\n");
    } else {
        if (ioctl(dsp_ctl_fd, OMAP_DSP_IOCTL_FBDIS) < 0) {
            perror("Xomap: ioctl OMAP_DSP_IOCTL_FBDIS failed");
        }
        close(dsp_ctl_fd);
    }
    rep.status = 1;

getout:
    if (client->swapped) {
        int n;

        swaps(&rep.sequenceNumber, n);
        swaps(&rep.status, n);
    }

    WriteToClient(client, sizeof(rep), (char *)&rep);
    return (client->noClientException);
}

static int SProcXSPCancelDSPArea(ClientPtr client)
{
    REQUEST(xXSPCancelDSPAreaReq);
    register int n;
    REQUEST_SIZE_MATCH(xXSPCancelDSPAreaReq);
    swaps(&stuff->screen_num, n);

    return ProcXSPCancelDSPArea(client);
}



static int ProcXSPSetPixelDoubling(ClientPtr client)
{
    REQUEST(xXSPSetPixelDoublingReq);
    xXSPSetPixelDoublingReply rep;
    REQUEST_SIZE_MATCH(xXSPSetPixelDoublingReq);
    xspScrPrivPtr pScrPriv;

    memset(&rep, 0, sizeof(rep));
    rep.type = X_Reply;
    rep.sequenceNumber = client->sequence;

    if (stuff->screen_num < 0 || stuff->screen_num > (screenInfo.numScreens - 1)) {
        rep.status = 0;
        goto getout;
    }

    pScrPriv = screenInfo.screens[stuff->screen_num]->devPrivates[xspScrPrivateIndex].ptr;

    if (!stuff->state) {
        pScrPriv->pixel_doubling = 0;
    } else {
        pScrPriv->pixel_doubling = 1;
    }

getout:
    if (client->swapped) {
        int n;

        swaps(&rep.sequenceNumber, n);
        swaps(&rep.status, n);
    }

    WriteToClient(client, sizeof(rep), (char *)&rep);
    return (client->noClientException);
}


static int SProcXSPSetPixelDoubling(ClientPtr client)
{
    REQUEST(xXSPSetPixelDoublingReq);
    register int n;
    REQUEST_SIZE_MATCH(xXSPSetPixelDoublingReq);
    swaps(&stuff->screen_num, n);
    swaps(&stuff->state, n);

    return ProcXSPSetPixelDoubling(client);
}



static void XSPClientCallback (CallbackListPtr *list,
        pointer closure,
        pointer data)
{
    NewClientInfoRec    *clientinfo = (NewClientInfoRec *) data;
    ClientPtr           pClient = clientinfo->client;
    xspScrPrivPtr pScrPriv;
    int s;

    if (clientinfo->setup != NULL) return;

    if (xsp_ts_client != NULL && xsp_ts_client == pClient) {

        /* Stop calibrating.  */
        xsp_ts_client = NULL;
        tslib_raw_event_hook = NULL;
        tslib_raw_event_closure = NULL;
    }

    /* cancel dsp area */
    for (s=0; s < screenInfo.numScreens; s++) {
        pScrPriv = screenInfo.screens[s]->devPrivates[xspScrPrivateIndex].ptr;

        if (pScrPriv->dsp_client != NULL 
                && pScrPriv->dsp_client == pClient) {

            pScrPriv->dsp_client = NULL;
            pScrPriv->dsp_enabled = 0;
            pScrPriv->dsp_box.x1 = 29999;
            pScrPriv->dsp_box.y1 = 29999;
            pScrPriv->dsp_box.x2 = -29999;
            pScrPriv->dsp_box.y2 = -29999;
        }
    }
}




static void ProcXSPCloseDown(ExtensionEntry *entry)
{
    int i;

    for (i=0; i < screenInfo.numScreens; i++) {
        xfree(screenInfo.screens[i]->devPrivates[xspScrPrivateIndex].ptr);
    }
    xspScrPrivateIndex = -1;
    XSPGeneration = -1;
    return;
}

#ifdef XFree86LOADER
static XF86ModuleVersionInfo XSPVersRec =
{
    XSP_EXTENSION_NAME,
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XF86_VERSION_CURRENT,
    XSP_EXTENSION_MAJOR, XSP_EXTENSION_MINOR, XSP_EXTENSION_PATCHLEVEL,
    ABI_CLASS_EXTENSION,
    ABI_EXTENSION_VERSION,
    MOD_CLASS_EXTENSION,
    {0,0,0,0}
};


static MODULESETUPPROTO(xspSetup);

XF86ModuleData xspModuleData = { &XSPVersRec, XSPSetup, NULL };

static pointer XSPSetup(pointer module, pointer opts, int *errmaj, int *errmin)
{
    XSPExtensionInit();

    return (pointer)TRUE;
}
#endif

